/* ================================================================
   routinen zum Scannen und Ueberpruefen von strings

   (c) Heisch Automation, Werner Heisch
       http://sites.inka.de/heisch

   ================================================================
   Stand  21.02.2009 
   Stand  29.01.2009 Start


!!! rules are not implemented yet !

Funktion:
Uebergebene Strings werden in token zerlegt, dabei werden die Separatoren 
unterschieden zwischen:
gsep = allgemeine Separatoren
        - allgemeine Separatoren werden nicht registriert,
        - am Stringanfang und Stringende gefundene allgemeine Separatoren 
          werden eliminiert
csep  = Separatoren, die aus einem Character bestehen

ssep  = Separatoren, die aus Strings bestehen.

Zur Verarbeitung:

Es muß eine apsan-structure mit absan_gen() erzeugt werden.
Alle weitere Operatioen laufen mit Hilfe dieser structure 

Der eingelesene Original-String wird in "obuf" gespeichert,
Beim Einlesen wird der String entsprechend ueber die 
Separatorenden definierten Regeln zerlegt. 
Bei der Zerlegung wird der Original-String auf weitere Buffer verteilt
und darauf werden pointer gesetzt, so dass token und Separatoren
am Ende als Strings vorliegen, die über pointer zugaenglich sind.

Token und Separatoren (csep und ssep, nicht gsep) werden abwechselnd in
buf[0] und buf[1] abgelegt. damit jeweils hinter einem Eintrag eine Luecke 
entsteht, um den String mit '\0" zu terminieren.


*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "apsan.h"
#include "apsan_intern.h"

/* ---------------------------------------------------- */
/* recalloKb allocates new and formerly used blocks like realloc(),
   but uses calloc style parameters.
   but the space is allocated in portions to 1 kbyte, to slower
   memory fragmentation.
   calloK never shortens an allocated memory block, if the desired
   size is shorter or equal to the just allocated, nothing is changed.

   !! calloKb() may be called previsiously, because nothings changes, al
      long as the formerly allocated memory size is sufficient.

   ptr : the formerly used pointer, or NULL, if first allocation
   len : on start: if ptr == NULL: don't care
                   id ptr != NULL: size of the formerly allocated memory
         on return: allocated memory in bytes;
   nelem  : count of requested elements of SIZE elsize
   elsize : size of one requested element in bytes

   RETURN VALUE 
   callocKb returns the pointer, or if failed, NULL.
   on success, len is returned to the lengt of the really allocated memory size.
   see: realloc(3)

   The returned memory block may be uninitialized.
*/

void *recalloKb (void *ptr,size_t *len,size_t nelem, size_t elsize)

{

  size_t request = nelem * elsize;

  if ((ptr != NULL) && ( *len >= request)) return ptr; /* memory is sufficient */
  /* adjust request to portions of 1 kb */
  request = (( (request - 1) >> 10) +1) << 10;
  if (( ptr = realloc(ptr,request)) == NULL) {
    *len = 0;
  } else {
    *len = request;
  }
  return ptr;
}

/* ---------------------------------------------------- */
/* free the space allocated for a csep definition */

void csep_free_mem( Apsan_set_type **typeset)

{
  Apsan_set_type *ts;

  if (*typeset == NULL)
    return ;
  ts = *typeset;
  if ( ts->csep != NULL) free( ts->csep);
  if ( ts->cptr != NULL) free( ts->cptr);
  ts->csep = NULL;
  ts->cptr = NULL;
}

/* ---------------------------------------------------- */
/* free the space allocated for a ssep definition */

void ssep_free_mem( Apsan_set_type **typeset)

{
  Apsan_set_type *ts;

  if (*typeset == NULL)
    return ;
  ts = *typeset;
  if ( ts->ssep != NULL) free( ts->ssep);
  if ( ts->sptr != NULL) free( ts->sptr);
  ts->ssep = NULL;
  ts->sptr = NULL;
  ts->smax = 0;       /* maximum count of strings ssep*/
  ts->ssep_space_max = 0 ; /* reserved space for strings */
  ts->scnt = 0;
}

/* ---------------------------------------------------- */
/* free the space allocated for a ssep definition */

void rules_free_mem( Apsan_set_type **typeset)

{
  Apsan_set_type *ts;

  if (*typeset == NULL)
    return ;
  ts = *typeset;
  if ( ts->rules != NULL) free( ts->rules);
  if ( ts->rptr != NULL) free( ts->rptr);
  ts->rules = NULL;
  ts->rptr = NULL;
  ts->rmax = 0;       /* maximum count of strings for rules */
  ts->rules_space_max = 0 ; /* reserved space for rules*/
  ts->rcnt = 0;
}

/* ---------------------------------------------------- */
/* free the space allocated for a type set */

void type_set_free_mem( Apsan_set_type **typeset)

{
  Apsan_set_type *ts;

  if (*typeset == NULL)
    return ;
  ts = *typeset;
  rules_free_mem (&ts);
  ssep_free_mem ( &ts);
  csep_free_mem ( &ts);
  free(ts);
  *typeset = NULL;
}


/* ---------------------------------------------------- */
/* create and initialize the data space for a type set,
      but not the space needed for cseps,sseps or rules 

   A type set will never been reallocated, with this function,
   therefore, if you want to re-use it, it has to be deleted with
   type_set_free_mem() before.

   but normally, this is not necessary, because just defined 
   type set can be redefined by redefining it's csep,ssep,rules.

   on return : 0 = successful,
              -1 = out of memory

*/

int type_set_gen( Apsan_set_type **typeset )

{
  Apsan_set_type *ts;
  size_t dummy;

// printf(" --- typeset_gen -- 1 ---\n"); fflush(stdout);

  if (( *typeset = ts = recalloKb(*typeset,&dummy,sizeof(Apsan_set_type),1)) == NULL)
    return -1;
// printf(" --- typeset_gen -- 2 ---\n"); fflush(stdout);
  ts->cptr = NULL;
  ts->csep = NULL;
  ts->cptr_space_max = 0;
  ts->csep_space_max = 0;
  ts->ccnt = 0;
  ts->sptr = NULL;
  ts->sptr_space_max = 0 ; /* reserved space for strings, */
  ts->smax = 0;       /* maximum count of strings ssep*/
  ts->ssep = NULL;
  ts->ssep_space_max = 0 ; /* reserved space for strings, */
  ts->scnt = 0;
  ts->rptr = NULL;
  ts->rptr_space_max = 0 ; /* reserved space for rules*/
  ts->rules = NULL;
  ts->rmax = 0;       /* maximum count of strings for rules */
  ts->rules_space_max = 0 ; /* reserved space for rules*/
  ts->rcnt = 0;
// printf(" --- typeset_gen -- 3 ---\n"); fflush(stdout);
  return 0;
}

/* ---------------------------------------------------- */
/* create and initialize the data space for a csep definition,

   If space has been allocated before  it will be used,
   if not sufficient or not allocated yet, it will be allocated.
   If the ponters are != NULL, it is assumed, that space is allocated. 

   on return : 0 = successful,
              -1 = out of memory

*/

int csep_gen( Apsan_set_type **typeset,size_t max_csep )

{
  Apsan_set_type *ts = *typeset;

  ts = *typeset;
// printf(" --- csep_gen -- 1 ---\n"); fflush(stdout);

//  ts->cmax = max_csep;
//  ts->ccnt = 0;
// printf(" --- csep_gen -- 2 ---\n"); fflush(stdout);

  if (
    (( ts->cptr = recalloKb(ts->cptr,&( ts->cptr_space_max),sizeof *ts->cptr,max_csep)) != NULL)
    && (( ts->csep = recalloKb(ts->csep,&( ts->csep_space_max),(max_csep + 1),1)) != NULL)
  ) { /* all buffers sucessfully allocated */
// printf(" --- csep_gen -- 3 ---\n"); fflush(stdout);
    ts->cmax = ts->cptr_space_max / sizeof (* ts->cptr) ;
    return 0;
  }
  /* no success, free all allocated memory, best in reverse order */
// printf(" --- csep_gen -- 4 ---\n"); fflush(stdout);

  csep_free_mem(&ts);
// printf(" --- csep_gen -- 5 ---\n"); fflush(stdout);

  return -1;
}

/* ---------------------------------------------------- */
/* create and initialize the data space for a ssep definition,

   If space has been allocated before  it will be used,
   if not sufficient or not allocated yet, it will be allocated.
   If the ponters are != NULL, it is assumed, that space is allocated. 

   on return : 0 = successful,
              -1 = out of memory

*/

int ssep_gen( Apsan_set_type **typeset,int max_ssep,int max_sspace)

{
  Apsan_set_type *ts = *typeset;

  ts->scnt = 0;

  if (
    (( ts->sptr = recalloKb(ts->sptr,&( ts->sptr_space_max),sizeof (Apsan_sreg_type),max_ssep)) != NULL)
    && (( ts->ssep = recalloKb(ts->ssep,&( ts->ssep_space_max),max_sspace,1)) != NULL)
  ) { /* all buffers sucessfully allocated */
    ts->smax = ts->sptr_space_max / sizeof (Apsan_sreg_type);
    return 0;
  }

  /* no success, free all allocated memory, best in reverse order */
  csep_free_mem(&ts);
  return -1;
}

/* ---------------------------------------------------- */
/* create and initialize the data space for rules definition,

   If space has been allocated before  it will be used,
   if not sufficient or not allocated yet, it will be allocated.
   If the ponters are != NULL, it is assumed, that space is allocated. 

   on return : 0 = successful,
              -1 = out of memory

*/

int rules_gen( Apsan_set_type **typeset,int max_rules,int max_rspace )

{
  Apsan_set_type *ts = *typeset;

  ts->rcnt = 0;

  if (
    (( ts->rptr = recalloKb(ts->rptr,&( ts->rptr_space_max),sizeof (Apsan_rule_type),max_rules)) != NULL)
    && (( ts->rules = recalloKb(ts->rules,&( ts->rules_space_max), ts->rules_space_max,1)) != NULL) /* reserved space for strings */
  ) { /* all buffers sucessfully allocated */
    ts->rmax = ts->rptr_space_max / sizeof (Apsan_rule_type);       /* maximum count of rules */
    return 0;
  }
  /* no success, free all allocated memory, best in reverse order */
  rules_free_mem(&ts);
  return -1;
}

/* ---------------------------------------------------- */
/* generate an new apsan-structure and initialize the data

Parameters : astru     = pointer to a apsan-structure
             data_len  = maximum length of data in bytes
             max_toks  = maximum count of tokens of the input line,
                        including possible separators CSEP and SSEP
             max_gsep  = maximum count of general separators
                        space will be allocated

             Also the first type set will be allocated and initialized:
             max_csep_0  = maximum count of character separators
                        space will be allocated
             max_ssep_0  = maximum count of string separators
                         only space for organisation will be allocated
             max_rules_0  = maximum count of rules
                         only space for organisation will be allocated

on return: 0 = structure is successfully initialized
               a pointer to the initialized structure is returned,
          -1 = no data left to initialize
               no pointer to structure is returned.

/* ------------ */
int apsan_gen(APSAN_P *astru,size_t data_len,int max_toks,\
    size_t max_gsep, size_t max_csep_0, int max_ssep_0, int max_rules_0)

{

  APSAN *a;
  size_t dummy;
  int i;

  if (( *astru = a = malloc(sizeof (APSAN))) == NULL) {
    return -1;
  }
  a->size_max = data_len;
  a->str_len = 0;
  a->obuf = NULL;
  a->buf[0] = NULL;
  a->buf[1] = NULL;
  a->reg = NULL;
  a->reg_max = max_toks;
  a->reg_cnt = 0;
  a->result = NULL;
  a->gsep = NULL;
  a->gmax =  max_gsep;
  a->gcnt = 0;
  a->active_set = 0;
  /* Ptrs to Type sets will be initialized to NULL */ 
  for (i=0;i<APSAN_MAX_SETS;i++) {
    a->s[i] = NULL;
  }
  if ( ((a->obuf = recalloKb(NULL,&(a->size_max),data_len,1)) != NULL) 
    && ((a->buf[0] = recalloKb(NULL,&dummy,data_len,1)) != NULL)
    && ((a->buf[1] = recalloKb(NULL,&dummy,data_len,1)) != NULL)
    && ((a->tbuf   = recalloKb(NULL,&dummy,data_len,1)) != NULL)
    && ((a->reg = recalloKb(NULL,&(a->reg_space_max),sizeof(Apsan_reg_type),max_toks)) != NULL)
    && ((a->result = recalloKb(NULL,&(dummy),a->reg_space_max + sizeof(int),1)) != NULL)
    && ((a->gsep = recalloKb(NULL,&(a->gmax),(max_gsep + 1),1)) != NULL)
    && ( type_set_gen(&(a->s[0])) == 0)
    && ( csep_gen( &(a->s[0]),max_csep_0 ) == 0)
    && ( ssep_gen( &(a->s[0]),max_ssep_0,max_ssep_0 *10) == 0)
    && ( rules_gen( &(a->s[0]),max_rules_0,max_rules_0 *10 ) == 0)
  ) { /* all buffers sucessfully allocated */
    return 0;
  }
   /* not enough memory for buffers and structures: destroy exit */
  type_set_free_mem(&(a->s[0]));
  if (a->gsep != NULL) free(a->gsep);
  if (a->result != NULL) free(a->result);
  if (a->reg != NULL) free(a->reg);
  if (a->tbuf != NULL) free(a->tbuf);
  if (a->buf[1] != NULL) free(a->buf[1]);
  if (a->buf[0] != NULL) free(a->buf[0]);
  if (a->obuf != NULL) free(a->obuf);
  free(a);
  return -1;
}


/* ---------------------------------------------------- */
/* close  a formally opened apsan structure

Parameters : astru     = pointer to a apsan-structure


on return: 0 = structure is successfully initialized
               a pointer to the initialized structure is returned,
          -1 = no data left to initialize
               no pointer to structure is returned.

/* ------------ */
void apsan_end(APSAN_P astru)

{
  int i;
  APSAN *a = astru;

  if ( astru == NULL) return;

  for (i = APSAN_MAX_SETS - 1 ; i >=0; i--) {
    type_set_free_mem(&(a->s[i]));
  }
  if (a->gsep != NULL) free(a->gsep);
  if (a->result != NULL) free(a->result);
  if (a->reg != NULL) free(a->reg);
  if (a->tbuf != NULL) free(a->tbuf);
  if (a->buf[1] != NULL) free(a->buf[1]);
  if (a->buf[0] != NULL) free(a->buf[0]);
  if (a->obuf != NULL) free(a->obuf);
  free(a);
}

/* ---------------------------------------------------- */
/* apsan_def() :define separators

     The SSEP and RULE separators are given in a single string, separated by ','
     id numbers are auto generated, beginning with 1.
   see also : apsan_def_id()


Parameters : astru     = pointer to a apsan-structure
                         (defined before with apsan_gen() )
             type      = APSAN_GSEP  global separators
                         APSAN_CSEP character separators
                         APSAN SSEP string separators
                         APSAN_RULE rule separators
             ts_no     = if type == APSAN_GSEP : dont care
                         else: number of the type set, to use.
                         range: 0 .. APSAN_MAX_SETS


             s         = string with definitions
                         The string separators are separated by ',', if ','
                         shall be treated as a sign, it shall be written twice; ',,'
                         APSAN RULE rule set
                         The rules are separated by ',', if ','
                         shall be treated as a sign, it shall be written twice; ',,'
                         rule syntax : 
                         Ending a token with '*' represents all signs, if '*' only shall 
                         be treated as a single char, it has to be written trice.

                         Special type checking tokens:
                         conditions: preceeding and following char must be a GSEP or CSEP 
                         :isint: checks a token against an integer including the
                                 optional preceeding sign
                         :isfloat: checks a token against a float including the
                                 optional preceeding sign.


on return: 0 = success
          -1 = no space left to insert s
          -2 = unknown type
          -3 = no space left to increase sptr -structure
          -4 = no space left to increase rule -structure
          -7 = number of typeset is out of range [0 ..APSAN_MAX_SETS] 
          -8 = no space left to genereate a new type set


/* ------------ */
int apsan_def(APSAN_P astru, int type,int ts_no, const char s[])

{
  APSAN *a = astru;
  int i ,j;
  int snr;
  int ret_val;

  size_t  size = strlen(s); /* read length of input string */

  if (type == APSAN_GSEP ) { /* global separators */
    if ((a->gsep = recalloKb(a->gsep,&(a->gmax),size,1)) == NULL)
        return -1;  /* no space left */
    strcpy((char *)(a->gsep),s); /* copy the new definitions */
    a->gcnt = size;
    return 0;
  }
  /* the number of a type set is requested : */
  if ((ts_no < 0) || (ts_no > APSAN_MAX_SETS)) {
    return -7; /* number of typeset is out of range */
  }

  a->active_set = ts_no; /* active set */ 

  /* check, if structure for the requested type set exists, else create it */
  if (a->s[ts_no] == NULL) { /* does not exist, create it */
    if (type_set_gen( &(a->s[ts_no])) != 0) { 
      return -8; /* no space left to genereate a new type set */
    }
  }
  /* ----- */
  if (type == APSAN_CSEP ) { /* character separators */
    if (ret_val = csep_gen(&(a->s[ts_no]),size) != 0) {
      return ret_val;
    }
    strcpy((a->s[ts_no])->csep,s); /* copy the new definitions */
    return 0;
  } else
  /* ----- */
  if (type == APSAN_SSEP ) { /* string separators */
    if (!size) { /* empty string */
      a->s[ts_no]->scnt = 0; /* no SSEPs */
      return 0;
    }
    snr = 1; /* at first, count how much separators strings */ 
    for (i = 0;i<size;i++) 
      if (s[i] ==',') snr ++;

    if (ret_val = ssep_gen(&(a->s[ts_no]),snr,size+2) != 0) {
      return ret_val;
    }
    strcpy((char *)a->s[ts_no]->ssep,s); /* copy the new definitions */
    a->s[ts_no]->ssep[size] =',';
    a->s[ts_no]->ssep[size+1] ='\0';

    /* seperate the strings, using ','als delimiter */
    /* first, break the strings into pieces and reduce ',,' to ',' */
    a->s[ts_no]->sptr[0].text = a->s[ts_no]->ssep; /* first entry */
    a->s[ts_no]->sptr[0].infinit = 0 ; /* preset */
    snr = 0;
    j = 0;
   for (i = 0;i<=size + 1;i++) {
      if (a->s[ts_no]->ssep[i] == ',') { /* delimiter found */
        if ((i <=size ) && (a->s[ts_no]->ssep[i+1] != ',')) { /* treat it as delimiter */
          a->s[ts_no]->ssep[i] = '\0'; /* End this string */
          if (a->s[ts_no]->ssep[i-1] == '*' ) { /* Wildcard ? */
            if (a->s[ts_no]->ssep[i-2] != '*' ) { /* Wildcard ! */
              a->s[ts_no]->sptr[snr].infinit = 1 ; /* set infinit */
            }
            a->s[ts_no]->ssep[i-1] = '\0'; /* anyway,delete one '*' */
          } 
          snr ++;
          a->s[ts_no]->sptr[snr].id = snr;
          a->s[ts_no]->sptr[snr].text = &(a->s[ts_no]->ssep[i+1]); /* beginning of the next */
        } else { /* treat is as one */
          a->s[ts_no]->ssep[j] = ','; /* ',' is a sign */
          if (i < size -1 ) i ++ ; /* jump over second ',' */
        }
      } else {
        a->s[ts_no]->ssep[j] = a->s[ts_no]->ssep[i];
      }
      j ++ ;
    }
    a->s[ts_no]->ssep[j] = '\0'; /* end of last string */
    a->s[ts_no]->scnt = snr;
    /* measuring the length of the read strings */
    for (i = 0;i<snr;i++) {
      a->s[ts_no]->sptr[i].len = strlen(a->s[ts_no]->sptr[i].text);
    }
    return 0;
  } else
  /* ----- */
  if (type == APSAN_RULE ) { /* rules */
    if (!size) { /* empty string */
      a->s[ts_no]->rcnt = 0; /* no rules */
      return 0;
    }
    /* first check the count of parts, */
    snr = 1;
    for (i = 0;i<size;i++) 
      if (s[i] ==',') snr ++;
    if (ret_val = rules_gen(&(a->s[ts_no]),snr,size) != 0) {
      return ret_val;
    }
    strcpy((char *)a->s[ts_no]->rules,s); /* copy the new definitions */
    a->s[ts_no]->rmax = snr + 1;
    /* seperate the strings, using ','als delimiter */
    /* first, break the strings into pieces and reduce ',,' to ',' */
    a->s[ts_no]->rptr[0].text = a->s[ts_no]->rules; /* first entry */
    snr = 1;
    j = 0;
    for (i = 0;i<size;i++) {
      if (a->s[ts_no]->rules[i] == ',') { /* delimiter found */
        if ((i < size -1 ) && (a->s[ts_no]->rules[i+1] != ',')) { /* treat it as delimiter */
          a->s[ts_no]->rules[i] = '\0'; /* End this string */
          a->s[ts_no]->rptr[snr].text = &(a->s[ts_no]->rules[i+1]); /* beginning of the next */
          a->s[ts_no]->rptr[snr].id = snr;
          snr ++;
        } else { /* treat is as one */
          a->s[ts_no]->rules[j] = ','; /* ',' is a sign */
          if (i < size -1 ) i ++ ; /* jump over second ',' */
        }
      } else {
        a->s[ts_no]->rules[j] = a->s[ts_no]->rules[i];
      }
      j ++ ;
    }
    a->s[ts_no]->rules[j] = '\0'; /* end of last string */
    a->s[ts_no]->rcnt = snr;
    /* measuring the length of the read strings */
    for (i = 0;i<snr;i++) {
      a->s[ts_no]->rptr[i].len = strlen(a->s[ts_no]->rptr[i].text);
    }
    return 0;
  }
  return -2; /* unknown command */
}


/* ---------------------------------------------------- */
/* define separators SSEP,RULES within a list with 
   user defined id  numbers.
This function aspects an array [1..cnt] of structures of the form a 
  struct { int id1, char[..]"name_1"
           int id2, char[..]"name_2"
         }
  where id must be from type integer 
  and the string can by of any lenght longer than 0, with any contents, ending
  with '\0'.

see also : apsan_def()
Using apsan_def(), the SSEP and RULE separators are given in a single string, 
separated by ','id numbers are auto generated, beginning with 1.

Using this function,the id numbers are defined individually.

If a formerly used definition will be overwritten, additionaly needed space will 
be attached, but remaining space will never be freed. 

Parameters : astru     = pointer to a apsan-structure
                         (defined before with apsan_gen() )
             type      = APSAN SSEP string separators
                         The string separators are separated by ',', if ','
                         shall be treated as a sign, it shall be written twice; ',,'
                         APSAN RULE rule set
                         The rules are separated by ',', if ','
                         shall be treated as a sign, it shall be written twice; ',,'
                         rule syntax : 
                         * represents all signs
             ts_no     = number of the type set, to use.
                         range: 0 .. APSAN_MAX_SETS

             s          = pointer to the structure of separators declared by 
                          Apsan_def_id
                         The string separators are separated by ',', if ','
                         shall be treated as a sign, it shall be written twice; ',,'
                         APSAN RULE rule set
                         The rules are separated by ',', if ','
                         shall be treated as a sign, it shall be written twice; ',,'
                         rule syntax : 
                         Ending a token with '*' represents all signs, if '*' only shall 
                         be treated as a single char, it has to be written trice.

                         Special type checking tokens:
                         conditions: preceeding and following char must be a GSEP or CSEP 
                         :isint: checks a token against an integer including the
                                 optional preceeding sign
                         :isfloat: checks a token against a float including the
                                 optional preceeding sign.


             cnt        = count of array elements



on return: 0 = success
          -1 = no space left to insert s
          -2 = unknown type
          -3 = no space left to increase sptr -structure
          -4 = no space left to increase rule -structure
          -7 = number of typeset is out of range [0 ..APSAN_MAX_SETS] 
          -8 = no space left to genereate a new type set
          -9 = this type of separators (GSEP,CSEP ) not allowed here.


/* ------------ */

int apsan_def_id(APSAN_P astru, int type,int ts_no, Apsan_def_id *s,int cnt)

{
  APSAN *a = astru;
  int i ,j;
  int d;
  size_t size;
  int ret_val;
  Apsan_def_id *s_ptr; 

  if ((type == APSAN_GSEP ) ||(type == APSAN_CSEP)) { /* global separators */

    return -9;
  }
  /* the number of a type set is requested : */
  if ((ts_no < 0) || (ts_no > APSAN_MAX_SETS)) {
    return -7; /* number of typeset is out of range */
  }
  /* check the real length of the needed space for string */
  s_ptr = s;
  size = 0;
  for (i=0;i<cnt;i++) {
    if ((j = strlen(s_ptr->str)) == 0) /* check, the size, if 0 then error */
      return -10;
    size += j;
    s_ptr ++ ;
  }
  /* 'size' contains  accumulated length of the string parts */

  a->active_set = ts_no; /* active set */ 
 
  /* check, if structure for the requested type set exists, else create it */
  if (a->s[ts_no] == NULL) { /* does not exist, create it */
    if (type_set_gen( &(a->s[ts_no])) != 0) { 
      return -8; /* no space left to genereate a new type set */
    }
  }
  /* ====== */
  if (type == APSAN_SSEP ) { /* string separators */
    if (!size) { /* empty string */
      a->s[ts_no]->scnt = 0; /* no SSEPs */
      return 0;
    }
    if (ret_val = ssep_gen(&(a->s[ts_no]),cnt,size) != 0) {
      return ret_val;
    }
    /* copy the new definitions */
    s_ptr = s;
    d = 0; /* index in ssep[] */
    for (i=0;i<cnt;i++) {
      a->s[ts_no]->sptr[i].infinit = 0;
      a->s[ts_no]->sptr[i].id = s_ptr->id; /* ID */
      j = strlen(s_ptr->str); /* including '\0'*/
      strncpy(&(a->s[ts_no]->ssep[d]),s_ptr->str,j+1); /* copy including '\0'*/
      a->s[ts_no]->sptr[i].text = &(a->s[ts_no]->ssep[d]); /* set the pointer to it */
      /* check, if infinite length of separator */
      if (a->s[ts_no]->sptr[i].text[j-1] == '*') { /* Last sign is '*' */
        if (a->s[ts_no]->sptr[i].text[j-2] != '*') { /* 2nd last also -> no infinit */
          a->s[ts_no]->sptr[i].infinit = 1;
        }
        /* anyway, delete one '*' */
        a->s[ts_no]->sptr[i].text[j-1] = '\0';
        j --;
      }
      a->s[ts_no]->sptr[i].len = j; /* register the length */
      s_ptr ++; /* hop to next source */
      d += + j +1; /* hop to next string destination */ 
    }
    a->s[ts_no]->scnt = cnt;
    return 0;
  } else
  /* ====== */
  if (type == APSAN_RULE ) { /* rules */
    if (!size) { /* empty string */
      a->s[ts_no]->rcnt = 0; /* no rules */
      return 0;
    }
    if (ret_val = rules_gen(&(a->s[ts_no]),cnt,size) != 0) {
      return ret_val;
    }
    /* copy the new definitions */
    s_ptr = s;
    d = 0; /* index in ssep[] */
    for (i=0;i<cnt;i++) {
      a->s[ts_no]->rptr[i].id = s_ptr->id ; /* ID */
      j = strlen(s_ptr->str) +1 ; /* including '\0'*/
      strncpy(&(a->s[ts_no]->rules[d]),s_ptr->str,j+1); /* copy to remaining memory */
      a->s[ts_no]->rptr[i].text = &(a->s[ts_no]->ssep[d]); /* set the pointer to it */
      a->s[ts_no]->rptr[i].len = j -1; /* register the lentgh */
      s_ptr ++; /* hop to next source */
      d += j; /* hop to next string destination */ 
    }
    a->s[ts_no]->rcnt = cnt;
    return 0;
  }
  return -2; /* unknown command */
}


/* ---------------------------------------------------- */
/* load a string and interpret
   or 
   new interpretation of the old string, maybe with a new types_set
   or new GSEP,CSEP,SSEP or RULES definitions

Parameters : astru     = pointer to a apsan-structure
                         (defined before with apsan_gen() )
             reg_what  = a mask, what to register and to count.
                         !!! tokens and defined GSEPs, CSEPs and SSEPs are always
                         !!! 'reg_what' only decides, if the are registered in reg. 
                         Possible values:
                         APSAN_NONE 0 = only token are registered, which are not matched by
                         separators 
                         if 'reg_what' is set to non zero ( not 0 or APSAN_NONE )
                         APSAN_TOKEN must be set for seperators and rules are 
                            part of the result.
                         APSAN_TOKEN 1 = only token
                         APSAN_CSEP  4 = token and cseps
                         APSAN_SSEP  8 = token and sseps
                         APSAN_RULE 16 = token and rule
                         APSAN_CSEP | APSAN_SSEP  12 = token,cseps and sseps
             ts_no     = number of the type set, to use.
                         if none, use -1 
                         else 0 .. APSAN_MAX_SETS

             s         = string to investigate
                         If set to NULL, the old string is evaluated again,
                         maybe with other SEPs od other 'reg_what'

on return: 0 >= success, count of found tokens and separators GSEP SSEP CSEP tand RULES ype
          -1 = buffers too short to insert s
               need higher "data_len" in "apsan_gen()".
          -5 = too much tokens, need higher "max_toks" in "apsan_gen()".
          -6 = requested type set is not defined
          -7 = number of typeset is out of range [-1 ..APSAN_MAX_SETS] 

Evaluation:
GSEPs are not registered in *reg.
First, the gseps at the end of the file are deleted, then the gseps 
at the start. 
Then the resulting inner part is evaluated:
  presedence: RULE-> GSEP -> SSEP -> CSEP
Subsequential GSEPs are registered as one, all others are registered
by their occurence.

/* ------------ */
int apsan_scan(APSAN_P astru,int reg_what,int ts_no, const char s[])

{
  APSAN *a = astru;

  size_t  size;
  char hstr[50];
  char *strt, *end;
  int i;
  int t_idx;
  int act_len;
  Apsan_set_type *tsp; /* ptr to active type set */
  int snr;
  int buf_nr;
  int reg_idx;
  int old_type;
  int run_idx,end_idx;
  signed char c, last_c;
  signed char found;
  signed char tok_chck  = 0;
  signed char csep_chck = 0;
  signed char ssep_chck = 0;
  signed char rule_chck = 0;
  signed char last_was_xsep;
  signed char this_is_xsep;

  if ( ((reg_what & APSAN_TOKEN) != 0 ) 
    || (!reg_what)) tok_chck = 1; /* register token in reg */
  if ((reg_what & APSAN_CSEP) != 0 ) csep_chck = 1; /* register CSEPs in reg */
  if ((reg_what & APSAN_SSEP) != 0 ) ssep_chck = 1; /* register SSEPs in reg */
  if ((reg_what & APSAN_RULE) != 0 ) rule_chck = 1; /* register RULEs in reg */
  if ( s != NULL) { /* new string */
    size = strlen(s); /* read length of input string */
    if (size >= a->size_max) return -1;
    strcpy((char *)(a->obuf),s); /* copy the original string */
    a->str_len = size;
  } else { /* check old string again */
    size = a->str_len;
  }

  /* GSEP may have been changed before, therefore anyway,find end and start */
  memset(a->tbuf,0,size +1);
  /* start the evaluation, first start */
  strt = a->obuf;
  end = &(a->obuf[size-1]);
  /* find the general separators at the end and cut them */
  do {
    found = 0;
    for(i=0;i<a->gcnt;i++) {
      if (*end == a->gsep[i]) {
        found = 1;
        end --;
        break;
      }
    }
  } while ((end > strt) && found);
  end ++;
  /* .. then the general separators at the beginning and cut them */
  do {
    found = 0;
    for(i=0;i<a->gcnt;i++)
      if (*strt == a->gsep[i]) {
        found = 1;
        strt ++;
        break;
      }
  } while ((end > strt) && found);

  /* evaluate the inner part */
  tsp = a->s[ts_no]; /* pointer to active type set */
  /* presedence: RULE -> GSEP -> SSEP -> CSEP */
  run_idx = strt - a->obuf;
  end_idx = end - a->obuf;
  /* set the tbuf with GSEP and CSEP to help SSEP */
  for (t_idx=run_idx;t_idx<=end_idx;t_idx++) {
    c = a->obuf[t_idx];
    found = 0;
    for(i=0;i<a->gcnt;i++) {
      if ( c == a->gsep[i]) {
        a->tbuf[t_idx] = APSAN_GSEP;
        found = 1;
        break;
      }
    }
    if (found == 0) {
      for(i=0;i< tsp->ccnt;i++) {
        if ( c == tsp->csep[i]) {
          a->tbuf[t_idx] = APSAN_CSEP;
          break;
        }
      }
    }
  }

  a->tbuf[end_idx+1] = APSAN_GSEP; /* security end */

  buf_nr  = 0;
  reg_idx = 0;
  (a->reg[reg_idx]).text = strt;
  (a->reg[reg_idx]).type = old_type = APSAN_NONE; /* up to now, no type */
  last_was_xsep = 1;
  do {
    /* first look for GSEPS */ 
    this_is_xsep = 0;
    found = 0;
    if (a->gcnt > 0) { /* search for gseps */
      for(i=0;i<a->gcnt;i++)
        if (a->obuf[run_idx] == a->gsep[i]) {
          found = 1;
          a->buf[0][run_idx] = '\0';
          a->buf[1][run_idx] = '\0';
          old_type = this_is_xsep = APSAN_GSEP;
          run_idx ++;
          break;
       }
    }
    if (!found && (tsp->scnt > 0) 
      && last_was_xsep )  { /* search for sseps */
      for(i=0;i<tsp->scnt;i++) {
        /* checking for special tokens */
        if (strcmp(tsp->sptr[i].text,":isint:") == 0) {
          act_len = 0;
          if ( ((c = a->obuf[run_idx]) == '+')
            || ( c == '-')
            || (( c >= '0') && (c <= '9')) ) {
            act_len = 1;
            last_c = c;
            while ( (( ( c = a->obuf[run_idx + act_len]) >= '0') && (c <= '9')) /* run through the digits */
              &&  (a->tbuf[run_idx + act_len] != APSAN_GSEP)
              && (a->tbuf[run_idx + act_len] != APSAN_CSEP)
             ) {
               act_len ++; last_c = c ;
             }
            if ( (( last_c >= '0') && (last_c <= '9'))    /* check the end */
              && ( act_len >= 1)
              && ( (a->tbuf[run_idx + act_len] == APSAN_GSEP)
                || (a->tbuf[run_idx + act_len] == APSAN_CSEP) )
            ) { 
              found = 1;
            }
          }
        } else if (strcmp(tsp->sptr[i].text,":isfloat:") == 0) {
          act_len = 0;
          if ( ((c =a->obuf[run_idx]) == '+')
            || ( c == '-')
            || (( c >= '0') && (c <= '9')) ) {
            act_len = 1; last_c = c ;
            while ( (( ( c = a->obuf[run_idx + act_len]) >= '0') && (c <= '9')) /* run through the digits */
              &&  (a->tbuf[run_idx + act_len] != APSAN_GSEP)
              && (a->tbuf[run_idx + act_len] != APSAN_CSEP)
             ) act_len ++;
            if ( ((c = a->obuf[run_idx + act_len]) == '.') /* period found ? */
             ) { 
               act_len ++ ;
               while ( (( ( c = a->obuf[run_idx + act_len]) >= '0') && (c <= '9')) /* run through the digits */
                 &&  (a->tbuf[run_idx + act_len] != APSAN_GSEP)
                 && (a->tbuf[run_idx + act_len] != APSAN_CSEP)
               ) {
                 act_len ++; last_c = c ;
               }
               if ( (( last_c >= '0') && (last_c <= '9'))    /* check the end */
                 && ( act_len >= 1)
                 && ( (a->tbuf[run_idx + act_len] == APSAN_GSEP)
                   || (a->tbuf[run_idx + act_len] == APSAN_CSEP) ) ) {
                 found = 1;
               }
            }
          }
        } else
        if (strncmp(&(a->obuf[run_idx]),tsp->sptr[i].text,tsp->sptr[i].len) == 0) {
          act_len = tsp->sptr[i].len;
          if (tsp->sptr[i].infinit) { /* match is infinit */
            while ( (a->tbuf[run_idx + act_len] != APSAN_GSEP)
              && (a->tbuf[run_idx + act_len] != APSAN_CSEP)
             ) {
               act_len ++ ;
             }
            found = 1;
          } else {
            if ( (a->tbuf[run_idx + act_len] == APSAN_GSEP)
              || (a->tbuf[run_idx + act_len] == APSAN_CSEP)
             )  {
              found = 1;
             }
          }
        }
        if (found) {
          a->buf[buf_nr][run_idx] = '\0'; /* terminate last string */
          buf_nr = (++ buf_nr) & 1; /* switch between 0 and 1 to other buffer */
          strncpy(&(a->buf[buf_nr][run_idx]),&(a->obuf[run_idx]),act_len);
          if (reg_idx >= a->reg_max - 1 ) return -5;
          if (ssep_chck) {
            a->reg[reg_idx].text = &(a->buf[buf_nr][run_idx]);
            a->reg[reg_idx].type = old_type = APSAN_SSEP;
            a->reg[reg_idx].id = tsp->sptr[i].id;
            a->reg[reg_idx].strt = run_idx; /* position in obuf */
            a->reg[reg_idx].end  = run_idx + act_len -1; /* end position in obuf */
            reg_idx ++ ;
          }
          run_idx += act_len;
          break;
        }
      }
    }
    if (!found && (tsp->ccnt > 0)) { /* search for cseps */
      for(i=0;i<tsp->ccnt;i++) {
        if (a->obuf[run_idx] == tsp->csep[i]) {
          found = 1;
          a->buf[buf_nr][run_idx] = '\0'; /* terminate last string */
          buf_nr = (++ buf_nr) & 1; /* switch between 0 and 1 to other buffer */
          a->buf[buf_nr][run_idx] = a->obuf[run_idx];
          if (reg_idx >= a->reg_max - 1) return -5; 
          if (csep_chck) {
            a->reg[reg_idx].text = &(a->buf[buf_nr][run_idx]);
            a->reg[reg_idx].type = APSAN_CSEP;
            a->reg[reg_idx].id = i;
            a->reg[reg_idx].strt = run_idx; /* position in obuf */
            a->reg[reg_idx].end  = run_idx; /* end position in obuf */
            reg_idx ++ ;
          }
          old_type = this_is_xsep = APSAN_CSEP;
          run_idx ++;
          break;
        }
      }
    }
    if (!found) { /* nothing found : this is a token */
      if (old_type != APSAN_TOKEN) { /* registration and buffer change */
        a->buf[buf_nr][run_idx] = '\0'; /* terminate last string */
        buf_nr = (++ buf_nr) & 1; /* switch between 0 and 1 to other buffer */
        if (reg_idx >= a->reg_max - 1 ) return -5; 
        if (tok_chck) {
          a->reg[reg_idx].text = &(a->buf[buf_nr][run_idx]);
          a->reg[reg_idx].type = APSAN_TOKEN;
          a->reg[reg_idx].id = 0;
          a->reg[reg_idx].strt = run_idx; /* position in obuf */
          a->reg[reg_idx].end  = 0;
          reg_idx ++ ;
        }
      }
      a->buf[buf_nr][run_idx] = a->obuf[run_idx];
      old_type = APSAN_TOKEN;
      run_idx ++;
    }
    last_was_xsep = this_is_xsep; /* to start the SSEP, RULE scann */
  } while (run_idx <= end_idx);
  a->buf[buf_nr][run_idx] = '\0'; /* terminate last string */
  a->reg_cnt = reg_idx;
  /* register the lengt of tokens */
  if (reg_idx) {
    for(i=0;i<reg_idx;i++)
      if (a->reg[i].type == APSAN_TOKEN)
        a->reg[i].end = a->reg[i].strt + strlen(a->reg[i].text) -1 ;
  }
  return reg_idx; /* count of found tokens and separators */
}


/* ---------------------------------------------------- */
/* report the result of last scan

Parameters : astru     = pointer to a apsan-structure
                         (defined before with apsan_gen() )

on return: NULL ==  no result found, no formaerly scan ?
                else : Ptr to structure Apsan_result

/* ------------ */
apsan_result_p apsan_result(APSAN_P astru)

{
  APSAN *a = astru;

  if ((a->result->cnt = a->reg_cnt) == 0) /* count of results */
    return (Apsan_result *)(a->result);

  /* copy the internal structures to avoid damage */
  memcpy (&(a->result->r),&(a->reg), sizeof (Apsan_reg_type) * (a->result->cnt));
  return a->result;
}

/* ---------------------------------------------------- */
/* report a string, part of the incoming line
   the resulting string is limited by '\0'

Parameters : astru     = pointer to a apsan-structure
                         (defined before with apsan_gen() )
             str       = string to be filled
             strt      = Index, start -position of the string
             end       = end position of the string 
                         OR 0 = return string up to the end last token != GSEP
                         OR -1 = return up to the complete endof input string,
                                 including all GSEP at the end of incoming string  
on return: (char *) to str

/* ------------ */
char *apsan_rstr(APSAN_P astru, char *str, int strt, int end)

{
  APSAN *a = astru;
  int h_end;

  if (end > 0) 
    h_end = end;
  else if (end == -1) 
    h_end = a->str_len;
  else
    h_end = a->reg[a->reg_cnt-1].end;

  strncpy(str,&(a->obuf[strt]),(h_end  - strt +1));
  str[h_end -strt +1] = '\0';

  return str;
}

/* ---------------------------------------------------- */
/* report a string, part of the incoming line
   start and end are defined by pointing to structure created by 
   the last apsan_scan() (and reported to the user by apsan_result().).
   the resulting string is limited by '\0'

   !! strt_idx and end_idx are starting at 0 !

Parameters : astru     = pointer to a apsan-structure
                         (defined before with apsan_gen() )
             str       = string to be filled

             strt_idx  = index of the token to be used as start:
                          0 .. n : nr of found token  according to 
                                   apsan_result()
             strt_how  =  0  = Start of next token behind this token
                          1  = including the token, pointed to
                          2  = next char behind the token, including GSEP + CSEP

             end_idx   =  index of the token to be used as End:
                          0 .. n : nr of found token according to 
                                   apsan_result()
             end_how   =  0  = End of last token before this token
                          1  = including the token, pointed to
                          2  = last char before the token including GSEP + CSEP


on return: (char *) to str
                    or NULL, if an error occured. 
                    The reason for the error is storted in a->errno.

/* ------------ */
char *apsan_rstrT(APSAN_P astru, char *str, 
                 int strt_idx, int strt_how,
                 int end_idx,  int end_how )

{
  APSAN *a = astru;
  size_t h_strt;
  size_t h_end;

  if ((strt_idx < 0) || (strt_idx >= a->reg_cnt)) { /* strt_idx out of range */
    a->errno = -11;
    return NULL;
  }
  if ((strt_how < 0) || (strt_how > 2)) { /* strt_how out of range */
    a->errno = -12;
    return NULL;
  }
  if ((end_idx < -1) || (end_idx >= a->reg_cnt)) { /* end_idx out of range */
    a->errno = -13;
    return NULL;
  }
  if ((end_how < 0) || (end_how > 2)) { /* end_how out of range */
    a->errno = -14;
    return NULL;
  }
  if (strt_idx > end_idx) { /* Start > end  */
    a->errno = -13;
    return NULL;
  }

  /* search the start position */
  switch(strt_how) {
    case 0 : if (strt_idx < (a->reg_cnt -1)) {
                h_strt = a->reg[strt_idx+1].strt;
             } else {
               a->errno = -11;
               return NULL;
             }
             break;
    case 1 : h_strt = a->reg[strt_idx].strt;
             break;
    case 2 : h_strt = a->reg[strt_idx].end + 1 ;
             break;
    default: break;
  }
  /* search the end position */
  switch(strt_how) {
    case 0 : if (end_idx > 0) {
                h_end = a->reg[end_idx-1].end;
             } else {
               a->errno = -13;
               return NULL;
             }
             break;
    case 1 : h_end = a->reg[end_idx].end;
             break;
    case 2 :  if (end_idx > 0) {
                h_end = a->reg[end_idx-1].strt - 1;
             } else {
               a->errno = -13;
               return NULL;
             }
             break;
    default: break;
  }
  return apsan_rstr(a,str,h_strt,h_end);
}

/* ---------------------------------------------------- */
/* report a token, number according to reported to the user 
   by apsan_result().).
   the resulting string is limited by '\0'

   !! strt_idx and end_idx are starting at 0 !

Parameters : astru     = pointer to a apsan-structure
                         (defined before with apsan_gen() )
             str       = string to be filled

             idx  = index of the token

on return: (char *) to str
                    oor NULL, if an error occured. 
                    The reason for the error is storted in a->errno

/* ------------ */
char *apsan_rTok (APSAN_P astru, char *str,int idx )

{
  APSAN *a = astru;
  size_t h_strt;
  size_t h_end;

  if ((idx < 0) || (idx >= a->reg_cnt)) { /* strt_idx out of range */
    a->errno = -15;
    return NULL;
  }

  h_strt = a->reg[idx].strt;
  h_end = a->reg[idx].end;
  return apsan_rstr(a,str,h_strt,h_end);
}


/* ---------------------------------------------------- */
/* print the structure contents
( for trouble shooting only)

/* ------------ */
void apsan_dump(APSAN_P astru)

{
  int i;
  int ts;

  APSAN *a = astru;

  printf("apsan_dump()..\n");fflush(stdout);

  printf("  apsan_ptr = %x\n",a); fflush(stdout);
  printf("  obuf     = %x",a->obuf); fflush(stdout);
  printf(" >>%s<<\n",a->obuf); fflush(stdout);
  printf("  size_max = %5i",a->size_max); fflush(stdout);
  printf("  str_len  = %5i\n",a->str_len); fflush(stdout);
  printf("  buf[0]   = %x",a->buf[0]); fflush(stdout);
  printf(" >>%s<<\n",a->buf[0]); fflush(stdout);
  printf("  buf[1]   = %x",a->buf[1]); fflush(stdout);
  printf(" >>%s<<\n",a->buf[1]); fflush(stdout);
  printf("  reg      = %x\n",a->reg); fflush(stdout);
  printf("  reg_space_max   = %i\n",a->reg_space_max); fflush(stdout);
  printf("  reg_max  = %i\n",a->reg_max); fflush(stdout);
  printf("  reg_cnt  = %i\n",a->reg_cnt); fflush(stdout);
  printf("  gsep     = %x",a->gsep); fflush(stdout);
  printf(" >>%s<<\n",a->gsep); fflush(stdout);
  printf("  gmax     = %i\n",a->gmax); fflush(stdout);
  printf("  gcnt     = %i\n",a->gcnt); fflush(stdout);
  /*  display type sets */
  for (ts= 0; ts<APSAN_MAX_SETS; ts++ ) {
    if (a->s[ts] != NULL ) {
      printf("--- type set %4i --------\n",ts); fflush(stdout);
      printf("  cptr     = %x\n",a->s[ts]->cptr); fflush(stdout);
      printf("  cptr_space_max   = %i\n",a->s[ts]->cptr_space_max); fflush(stdout);
      printf("  csep     = %x",a->s[ts]->csep); fflush(stdout);
      printf(" >>%s<<\n",a->s[ts]->csep); fflush(stdout);
      printf("  cmax     = %i\n",a->s[ts]->cmax); fflush(stdout);
      printf("  ccnt     = %i\n",a->s[ts]->ccnt); fflush(stdout);
      printf("  sptr     = %x\n",a->s[ts]->sptr); fflush(stdout);
      printf("  ssep     = %x\n",a->s[ts]->ssep); fflush(stdout);
      printf("  smax     = %i\n",a->s[ts]->smax); fflush(stdout);
      printf("  ssep_space_max   = %i\n",a->s[ts]->ssep_space_max); fflush(stdout);
      printf("  scnt     = %i\n",a->s[ts]->scnt); fflush(stdout);
      for (i=0;i<a->s[ts]->scnt;i++) {
        printf("    str[%2i]:>>%s<<\n",i,a->s[ts]->sptr[i].text); fflush(stdout); 
      }
      printf("  rptr     = %x\n",a->s[ts]->rptr); fflush(stdout);
      printf("  rptr_space_max   = %i\n",a->s[ts]->rptr_space_max); fflush(stdout);
      printf("  rules     = %x\n",a->s[ts]->rules); fflush(stdout);
      printf("  rmax     = %i\n",a->s[ts]->rmax); fflush(stdout);
      printf("  rules_space_max   = %i\n",a->s[ts]->rules_space_max); fflush(stdout);
      printf("  rcnt     = %i\n",a->s[ts]->rcnt); fflush(stdout);
      for (i=0;i<a->s[ts]->rcnt;i++) {
        printf("    str[%2i]:>>%s<<\n",i,a->s[ts]->rptr[i].text); fflush(stdout); 
      }
    }
  }
}

/* ---------------------------------------------------- */
/* print the result of the last scan */

void apsan_print_result( APSAN_P s)

{
   int i;
   apsan_result_p r = apsan_result(s);

  printf("---- RESULT found %4i tokens ---------\n",r->cnt);
  for(i = 0; i < (r->cnt); i++) {
    printf("nr = %3i, type = %i , id = %5i, strt = %4i end = %4i >>%s<<\n",
         i,r->r[i].type,r->r[i].id,r->r[i].strt,r->r[i].end,r->r[i].text);
  }
  printf("---- RESULT END -----------------------\n");
}



