/* ================================================================
   routinen zum Scannen und Ueberpruefen von strings

   (c) Heisch Automation, Werner Heisch
       http://sites.inka.de/heisch

   ================================================================
   Stand  29.01.2009 Start

Funktion:
Uebergebene Strings werden in token zerlegt, dabei werden die Separatoren 
unterschieden zwischen:
gsep = allgemeine Separatoren
        - allgemeine Separatoren werden nicht registriert,
        - am Stringanfang und Stringende gefundene allgemeine Separatoren 
          werden eliminiert
csep  = Separatoren, die aus einem Character bestehen

ssep  = Separatoren, die aus Strings bestehen.
        Spezielles: Wildcard Endung
        Wenn der Separator mit '*' endet sind alle Worte mit gleichem Anfang
        ein Treffer. Das gesuchte Wort endet dann am nächsten GSEP oder CSEP
        Soll ein Wort exakt treffen, der letzt Buchstabe ist aber ein '*', so muss
        als Separator-Ende '**' geschrieben werden.
        Wenn exakt getroffen werden soll, muss das dem gefundenen Wort folgende
        Zeichen ein GSEP oder CSEP sein.

rule  = derzeit noch nicht realisiert.

Zur Verarbeitung:

Es muß eine apsan-structure mit absan_gen() erzeugt werden.
Alle weitere Operatioen laufen mit Hilfe dieser structure 

Der eingelesene Original-String wird in "obuf" gespeichert,
Beim Einlesen wird der String entsprechend ueber die 
Separatorenden definierten Regeln zerlegt. 
Bei der Zerlegung wird der Original-String auf weitere Buffer verteilt
und darauf werden pointer gesetzt, so dass token und Separatoren
am Ende als Strings vorliegen, die über pointer zugaenglich sind.

Token und Separatoren (csep und ssep, nicht gsep) werden abwechselnd in
buf[0] und buf[1] abgelegt. damit jeweils hinter einem Eintrag eine Luecke 
entsteht, um den String mit '\0" zu terminieren.

*/

/* ---------------- LIST OF ERROR CODES --------------- */
/*
  0 = success
 -1 = no space left to insert s
 -2 = unknown type
 -3 = no space left to increase sptr -structure
 -4 = no space left to increase rule -structure
 -5 = too much tokens, need higher "max_toks" in "apsan_gen()".
 -6 = requested type set is not defined
 -7 = number of typeset is out of range [0 ..APSAN_MAX_SETS] 
 -8 = no space left to genereate a new type set
 -9 = this type of separators (GSEP,CSEP) not allowed here.
-10 = size of seperator definition = 0 

*/


#include<sys/types.h>






/* ---------------------------------------------------- */
/* declarations of the data-structures for APSAN          */

/* this structure holds internal informations about string separators */
typedef struct apsan_sreg_type {
  char *text;             /* Pointer to the separator string */
  signed char infinit;          /* if serarator ends with *, then 1 else 0 */
  size_t len;             /* Length of the ssep */
  int id;                 /* for lexical scans */
} Apsan_sreg_type;

/*  --- managing rules ------------- */

typedef struct apsan_match_set_type {
  int type;               /* 0 = Char, stored in 'c'
                             1 = string, set_ptr ->
                             2 = pattern set_ptr -> */
  signed char  c;                /* char if type == 'set' */
  int ocur_min;           /* minimum count of occurences */
  int ocur_max;           /* maximum count of occurences */
  int occupation;         /* count of bytes in  souce buffer 
                             to be occupied by this pattern/string/byte */
  size_t set_len;         /* length of the string of the following set */
  long alpha_vect;        /* vector, containing :ISALNUM: :ISALPHA: etc */
  char  *set_ptr;         /* pointer to a set, if type = set */
  int akt_strt;           /* actual start index */
  int akt_end;            /* actual end index */
} Apsan_match_set_type;

typedef struct apsan_rule_type {
  char *text;             /* Pointer to the rule string */
  size_t len;             /* Length of the rule string */
  int id;                 /* for lexical scans */
  int m_cnt;              /* actual count of m entrys */
  Apsan_match_set_type m[APSAN_MAXPATTERN]; /* array of pattern-match methods */
} Apsan_rule_type;

/*  --- structure of a set type ------------- */

typedef struct apsan_set_type {
  char **cptr;             /* pointer to list of cseps */
  size_t cptr_space_max;   /* size of actually allocated  memory for *cptr */
  int cmax;                /* maximum count of strings csep */
  char *csep;              /* list of cseps */
  size_t csep_space_max;   /* reserved space for all csep */
  int ccnt;		   /* count of csep */

  Apsan_sreg_type *sptr;   /* pointer to list of sseps */
  size_t sptr_space_max;   /* size of actually allocated  memory for *sptr */
  int smax;                /* maximum count of strings ssep */
  char *ssep;              /* list of sseps */
  size_t ssep_space_max;   /* reserved space for all ssep */
  int scnt;		   /* count of strings ssep */

  Apsan_rule_type *rptr;   /* pointer to list of rules */
  size_t rptr_space_max;   /* size of actually allocated  memory for *sptr */
  int rmax;                /* maximum count of strings rules */
  char *rules;             /* list of rules */
  size_t rules_space_max;  /* reserved space for all rule strings */
  int rcnt;		   /* count of rules strings  */
} Apsan_set_type;

/* ------------- the APSAN main structure --------------- */

typedef struct apsan_stru {
  size_t  size_max;        /* maximum length, length of the created buffers */
  size_t str_len;          /* actual length of the evaluated string */
  char *obuf;              /* buffer with the original string */
  char *buf[2];            /* buffers with tokens and separators */
  char *tbuf;              /* control buffer. The type of fond matches is written here */
  Apsan_reg_type *reg;     /* pointer to the registration */
  size_t reg_space_max;    /* size of actually allocated  memory for reg */
  int reg_max;		   /* maximum count of possible tokens and separators */
  int reg_cnt;		   /* count of tokens and separators */
  Apsan_result *result;  /* structure for presenting the results externally,
                              same dimensions like Apsan_reg_type + size(int) */
  char *gsep;              /* list of gseps */
  size_t gmax;		   /* maximum count of gsep */
  int gcnt; 		   /* count of gsep */
  int active_set;	   /* active set, set by  last loaded , selected etc */
  Apsan_set_type *s [APSAN_MAX_SETS+1]; /* up to .. sets of typesete */
  int errno;               /* error number, set by functions without a error
                              return, (i.e. string returning funktions( only overwritten,
                              never deleted */
} APSAN;


